
/*
json-tls	v1.1.1
author: 郭斌勇
license: ISC
homepage: https://github.com/GuoBinyong/json-tls#readme
repository: https://github.com/GuoBinyong/json-tls
description: json-tls 提供了JSON相关的工具，如:customJSONStringify、customJSONParse、deepCopyByJSON 等
*/

'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var typeTls = require('type-tls');
var typeReviver = require('type-reviver');
var comTools = require('com-tools');

//ReviverPair 的类型守卫
function isReviverPair(target) {
    return target && typeof target.string === "function" && typeof target.parse === "function";
}

/**
 * 安全地解析字符串，不会抛出错误，返回一个解析结果的信息对象
 * @param text : string   必需， 一个有效的 JSON 字符串。
 * @param reviver ?: function   可选，一个转换结果的函数， 将为对象的每个成员调用此函数。
 * @returns {parsed: boolean, result: string | JSONObject }   解析的结果； parsed ：表示是否成功解析； result ： 最终解析的结果，如果成功解析，则该值为解析后的JSON对象，如果未成功解析，则该值为原字符串 text
 */
function safelyParse(text, reviver) {
    let parseInfo = {
        parsed: true,
        result: text
    };
    try {
        parseInfo.result = JSON.parse(text, reviver);
    }
    catch (e) {
        parseInfo.parsed = false;
    }
    return parseInfo;
}
function deepStringify(target, replacer, space) {
    let newTarget = Object.keys(target).reduce((total, key) => {
        let prop = target[key];
        if (typeof prop == "object") {
            prop = deepStringify(prop, replacer, space);
        }
        total[key] = prop;
        return total;
    }, {});
    // @ts-ignore
    return JSON.stringify(newTarget, replacer, space);
}
/**
 * 深度解析； deepStringify 的逆运算
 * @param text
 * @param reviver
 * @returns any
 */
function deepParse(text, reviver) {
    let result = correctParse(text, reviver);
    if (typeof result == "object") {
        Object.keys(result).forEach((key) => {
            let prop = result[key];
            prop = deepParse(prop, reviver);
            result[key] = prop;
        });
    }
    return result;
}
/**
 * 判断 text 是否是有效的JSON字符串
 * @returns boolean
 */
function isJSONString(text) {
    try {
        var parsed = JSON.parse(text);
        var isJSON = parsed !== text;
    }
    catch (e) {
        isJSON = false;
    }
    return isJSON;
}
/**
 * correctParse(text,reviver)
 * 正确地解析，即：如果 text 是有效的 JSON 字符串，则解析它并返回解析后的结果，否则，返回原来的text
 * @param text ?:  string   被解析的字符串
 * @returns any
 */
function correctParse(text, reviver) {
    try {
        var parsed = JSON.parse(text, reviver);
        var isJSON = parsed !== text;
    }
    catch (e) {
        isJSON = false;
    }
    if (!isJSON) {
        parsed = text;
    }
    return parsed;
}

// Date ----------
//原始的 toJSON 方法
const Date_toJSON = Date.prototype.toJSON;
//StringifyReviver
function Date_StringifyReviver(key, value, type, callCount, stringifyOptions) {
    if (stringifyOptions.skip || stringifyOptions.skipMark || (callCount === 1 && stringifyOptions.skipRootMark)) {
        return Date_toJSON ? Date_toJSON.call(value) : value.toString();
    }
    return value.valueOf();
}
//ParseReviver
function Date_ParseReviver(key, value, type, callCount) {
    return new Date(value);
}
// Map ----------
//原始的 toJSON 方法
// @ts-ignore
const Map_toJSON = Map.prototype.toJSON;
//StringifyReviver
function Map_StringifyReviver(key, value, type, callCount, stringifyOptions) {
    if (Map_toJSON && (stringifyOptions.skip || stringifyOptions.skipMark || (callCount === 1 && stringifyOptions.skipRootMark))) {
        return Map_toJSON.call(value);
    }
    return Array.from(value);
}
//ParseReviver
function Map_ParseReviver(key, value, type, callCount) {
    return new Map(value);
}
// Set ----------
//原始的 toJSON 方法
// @ts-ignore
const Set_toJSON = Set.prototype.toJSON;
//StringifyReviver
function Set_StringifyReviver(key, value, type, callCount, stringifyOptions) {
    if (Set_toJSON && (stringifyOptions.skip || stringifyOptions.skipMark || (callCount === 1 && stringifyOptions.skipRootMark))) {
        return Set_toJSON.call(value);
    }
    return Array.from(value);
}
//ParseReviver
function Set_ParseReviver(key, value, type, callCount) {
    return new Set(value);
}
// URL ----------
//原始的 toJSON 方法
// @ts-ignore
const URL_toJSON = URL.prototype.toJSON;
//StringifyReviver
function URL_StringifyReviver(key, value, type, callCount, stringifyOptions) {
    if (URL_toJSON && (stringifyOptions.skip || stringifyOptions.skipMark || (callCount === 1 && stringifyOptions.skipRootMark))) {
        return URL_toJSON.call(value);
    }
    return value.href;
}
//ParseReviver
function URL_ParseReviver(key, value, type, callCount) {
    return new URL(value);
}
// RegExp ----------
//原始的 toJSON 方法
// @ts-ignore
const RegExp_toJSON = RegExp.prototype.toJSON;
//StringifyReviver
function RegExp_StringifyReviver(key, value, type, callCount, stringifyOptions) {
    if (stringifyOptions.skip || stringifyOptions.skipMark || (callCount === 1 && stringifyOptions.skipRootMark)) {
        return RegExp_toJSON ? RegExp_toJSON.call(value) : value.toString();
    }
    return {
        source: value.source,
        flags: value.flags
    };
}
//ParseReviver
function RegExp_ParseReviver(key, value, type, callCount) {
    let { source, flags } = value;
    let reg = new RegExp(source, flags);
    return reg;
}
// Function ----------
//原始的 toJSON 方法
// @ts-ignore
const Function_toJSON = Function.prototype.toJSON;
//StringifyReviver
function Function_StringifyReviver(key, value, type, callCount, stringifyOptions) {
    if (Function_toJSON && stringifyOptions.skip || stringifyOptions.skipMark || (callCount === 1 && stringifyOptions.skipRootMark)) {
        return Function_toJSON.call(value);
    }
    return {
        source: value.toString(),
        name: value.name
    };
}
//ParseReviver
function Function_ParseReviver(key, value, type, callCount) {
    /*
    考虑到安全和性能的原因，弃用 eval，改用 Function；
    */
    return comTools.createFunctionBy(value.source, value.name, true);
    /*
    匿名函数 不能用于 函数声明，只能用于函数表达式；
    所以，如果 value 是匿名函数，则 `eval(value);` 会报错： `Uncaught SyntaxError: Function statements require a function name`
    此时，只需要用 `()` 将其转为 函数表达式 即可；
    */
    // return eval(`(${value})`);
}
const typeReviverArray = [
    [Date, { string: Date_StringifyReviver, parse: Date_ParseReviver }],
    [Map, { string: Map_StringifyReviver, parse: Map_ParseReviver }],
    [Set, { string: Set_StringifyReviver, parse: Set_ParseReviver }],
    [URL, { string: URL_StringifyReviver, parse: URL_ParseReviver }],
    [RegExp, { string: RegExp_StringifyReviver, parse: RegExp_ParseReviver }],
    [Function, { string: Function_StringifyReviver, parse: Function_ParseReviver }],
];

/**
 * 默认的 Mark 值
 */
const _defaultMark = "__MarKOfCustomJSON__";
/**
 * 默认 Revier 的 TypeName
 */
const typeNameOfDefaultRevier = "default";
function createCustomJSONStringify(presetTypeReviverMap) {
    /**
     * 自定义JSON序列化；可根据数据类型来自定义序列化方案； 给合  customJSONParse  可实现对任意类型的数据 进行序列化 并完整（无信息丢失）还原；
     * @param value: any   被序列化的对象
     * @param typeRevivers?:TypeRevivers|null    定义 类型 与 其对类的 自定义序列化函数；
     * @param options:JSONStringifyOptions    选项
     */
    function customJSONStringify(value, typeRevivers, options = {}) {
        try {
            let presetTRMap = customJSONStringify.presetTypeReviverMap;
            if (typeRevivers || presetTRMap.size > 0) {
                let mergedTRArr = typeReviver.mergeTypeRevivers(presetTRMap, typeRevivers);
                var parseInfo = typeReviver.flatParseTypeRevivers(mergedTRArr);
                let disDefaultArr = parseInfo.typeFun;
                //禁用toJSON
                if (disDefaultArr.length > 0) {
                    var disTypeInfos = disDefaultArr.reduce(function (infoArr, fun) {
                        let toJSON = fun.prototype.toJSON;
                        if (toJSON) {
                            fun.prototype.toJSON = undefined;
                            infoArr.push({
                                type: fun,
                                toJSON: toJSON
                            });
                        }
                        return infoArr;
                    }, []);
                }
                var trObj = parseInfo.trObject;
                if (trObj) {
                    let opts = Object.assign({}, options);
                    let mark = opts.mark;
                    mark = opts.mark = mark == null ? _defaultMark : mark;
                    let markType = mark + "Type";
                    let markValue = mark + "Value";
                    let callCount = 0; // stringifyReviver 的调用次数
                    const defaultTypeName = typeNameOfDefaultRevier;
                    const defaultRevier = trObj[defaultTypeName];
                    const defaultRevierFun = (defaultRevier && typeof defaultRevier === "object") ? defaultRevier.string : defaultRevier;
                    let stringifyReviver = function (key, value) {
                        ++callCount;
                        //此处的写法没有问题，这样写是为了提高性能
                        switch (key) {
                            case mark: ;
                            case markType: ;
                            case markValue: return value;
                        }
                        if (opts.skipRoot && callCount === 1) { //需要跳过这次处理
                            return value;
                        }
                        let typeName = typeTls.getExactTypeNameOf(value);
                        let revier = trObj[typeName];
                        let revierFun = (revier && typeof revier === "object") ? revier.string : revier;
                        if (!revierFun) {
                            if (defaultRevierFun) {
                                typeName = defaultTypeName;
                                revier = defaultRevier;
                                revierFun = defaultRevierFun;
                            }
                            else {
                                return value;
                            }
                        }
                        let rerOpts = Object.assign({}, opts);
                        let rerRes = revierFun.call(this, key, value, typeName, callCount, rerOpts);
                        if (rerOpts.skip) { //需要放在上一句 `revier.call(this,key,value,typeName,callCount,rerOpts)` 的后面，因为 revier 可修改 rerOpts 的值
                            return value;
                        }
                        /*
                        在以下任一情况下，均不会添加 mark
                        - revier 返回 undefined  :  `rerRes === undefined`
                        - skipMark 为 true : `rerOpts.skipMark`
                        - value 是被 customJSONStringify 最初序列化的目标（即：根） 且  skipRootMark 为 true : `rerOpts.skipRootMark && callCount === 1`
                        */
                        if (rerRes === undefined || rerOpts.skipMark || (rerOpts.skipRootMark && callCount === 1)) {
                            return rerRes;
                        }
                        return {
                            [mark]: true,
                            [markType]: typeName,
                            [markValue]: rerRes
                        };
                    };
                    var jsonStr = JSON.stringify(value, stringifyReviver, opts.space);
                }
            }
            // @ts-ignore
            if (!trObj) {
                jsonStr = JSON.stringify(value, null, options.space);
            }
        }
        catch (e) {
            throw e;
        }
        finally {
            //取消禁用toJSON
            if (disTypeInfos && disTypeInfos.length > 0) {
                disTypeInfos.forEach(function (funInfo) {
                    funInfo.type.prototype.toJSON = funInfo.toJSON;
                });
            }
            // @ts-ignore
            return jsonStr;
        }
    }
    Object.defineProperty(customJSONStringify, "presetTypeReviverMap", {
        configurable: true,
        enumerable: true,
        get: function () {
            if (!this._presetTypeReviverMap) {
                this._presetTypeReviverMap = new Map();
            }
            return this._presetTypeReviverMap;
        },
        set: function (newValue) {
            if (newValue instanceof Map) {
                this._presetTypeReviverMap = newValue;
            }
        }
    });
    if (presetTypeReviverMap) {
        customJSONStringify.presetTypeReviverMap = presetTypeReviverMap;
    }
    return customJSONStringify;
}
(function (LostRevier) {
    LostRevier["original"] = "original";
    LostRevier["parse"] = "parse";
    LostRevier["ignore"] = "ignore";
})(exports.LostRevier || (exports.LostRevier = {}));
function createCustomJSONParse(presetTypeReviverMap) {
    /**
     * 自定义JSON解析；可根据类型自定义解析逻辑；
     * @param text
     * @param typeRevivers
     * @param options
     */
    function customJSONParse(text, typeRevivers, options = {}) {
        let lostRevier = options.lostRevier || exports.LostRevier.parse;
        let presetTRMap = customJSONParse.presetTypeReviverMap;
        if (typeRevivers || presetTRMap.size > 0) {
            let mergedTRArr = typeReviver.mergeTypeRevivers(presetTRMap, typeRevivers);
            let parseInfo = typeReviver.flatParseTypeRevivers(mergedTRArr);
            var trObj = parseInfo.trObject;
        }
        // @ts-ignore
        if (!trObj && lostRevier === exports.LostRevier.original) {
            return JSON.parse(text);
        }
        let mark = options.mark == null ? _defaultMark : options.mark;
        let markType = mark + "Type";
        let markValue = mark + "Value";
        let callCount = 0; // parseReviver 的调用次数
        function parseReviver(key, value) {
            ++callCount;
            if (typeTls.isBaseType(value)) {
                return value;
            }
            let typeName = value[markType];
            if (!(typeName && value[mark])) {
                return value;
            }
            var realValue = value[markValue];
            let revier = trObj[typeName];
            let revierFun = (revier && typeof revier === "object") ? revier.parse : revier;
            if (!revierFun) {
                switch (lostRevier) {
                    case exports.LostRevier.ignore: return undefined;
                    case exports.LostRevier.original: return value;
                    default: return realValue;
                }
            }
            return revierFun.call(this, key, realValue, typeName, callCount);
        }
        return JSON.parse(text, parseReviver);
    }
    Object.defineProperty(customJSONParse, "presetTypeReviverMap", {
        configurable: true,
        enumerable: true,
        get: function () {
            if (!this._presetTypeReviverMap) {
                this._presetTypeReviverMap = new Map();
            }
            return this._presetTypeReviverMap;
        },
        set: function (newValue) {
            if (newValue instanceof Map) {
                this._presetTypeReviverMap = newValue;
            }
        }
    });
    if (presetTypeReviverMap) {
        customJSONParse.presetTypeReviverMap = presetTypeReviverMap;
    }
    return customJSONParse;
}
function createCustomJSON(presetTypeReviverMap) {
    return {
        stringify: createCustomJSONStringify(presetTypeReviverMap),
        parse: createCustomJSONParse(presetTypeReviverMap)
    };
}
/**
 * 默认预置的TypeReviverMap
 */
const defaultPresetTypeReviverMap = new Map(typeReviverArray);
/**
 * 自定义JSON序列化；可根据数据类型来自定义序列化方案； 给合  customJSONParse  可实现对任意类型的数据 进行序列化 并完整（无信息丢失）还原；
 * @param value: any   被序列化的对象
 * @param typeRevivers?:TypeRevivers|null    定义 类型 与 其对类的 自定义序列化函数；
 * @param options:JSONStringifyOptions    选项
 */
const customJSONStringify = createCustomJSONStringify(defaultPresetTypeReviverMap);
/**
 * 自定义JSON解析；可根据类型自定义解析逻辑；
 * @param text
 * @param typeRevivers
 * @param options
 */
const customJSONParse = createCustomJSONParse(defaultPresetTypeReviverMap);

/**
 * StringParseTypeRevivers 的类型守卫
 * @param target
 */
function isTypeReviversPair(target) {
    return target && typeof target === "object" && !Array.isArray(target) && (target.string || target.parse);
}
/**
 * 创建纯净的 customJSON
 */
const customJSON = createCustomJSON();
/**
 * 创建带 presetTypeReviverMap 的 deepCopyByJSON 的工厂函数
 * @param presetTypeReviverMap
 */
function createDeepCopyByJSONWith(presetTypeReviverMap) {
    /**
     * 通过 自定义 JSON 序列人的方式 对 value 进行 深拷贝
     * @param value
     * @param dcTypeRevivers
     */
    function deepCopyByJSON(value, typeReviversOpts) {
        if (typeTls.isBaseType(value)) {
            return value;
        }
        let preTRArr = typeReviver.toTypeReviverArray(deepCopyByJSON.presetTypeReviverMap);
        let strAllTRArr = preTRArr;
        let parseAllTRArr = preTRArr;
        if (typeReviversOpts) {
            if (isTypeReviversPair(typeReviversOpts)) {
                let strTRs = typeReviversOpts.string;
                if (strTRs) {
                    let strTRArr = typeReviver.toTypeReviverArray(strTRs);
                    strAllTRArr = strAllTRArr.concat(strTRArr);
                }
                let parseTRs = typeReviversOpts.parse;
                if (parseTRs) {
                    let parseTRArr = typeReviver.toTypeReviverArray(parseTRs);
                    parseAllTRArr = parseAllTRArr.concat(parseTRArr);
                }
            }
            else {
                let trArr = typeReviver.toTypeReviverArray(typeReviversOpts);
                strAllTRArr = strAllTRArr.concat(trArr);
                parseAllTRArr = strAllTRArr;
            }
        }
        var str = customJSON.stringify(value, strAllTRArr);
        return customJSON.parse(str, parseAllTRArr);
    }
    Object.defineProperty(deepCopyByJSON, "presetTypeReviverMap", {
        configurable: true,
        enumerable: true,
        get: function () {
            if (!this._presetTypeReviverMap) {
                this._presetTypeReviverMap = new Map();
            }
            return this._presetTypeReviverMap;
        },
        set: function (newValue) {
            if (newValue instanceof Map) {
                this._presetTypeReviverMap = newValue;
            }
        }
    });
    if (presetTypeReviverMap) {
        deepCopyByJSON.presetTypeReviverMap = presetTypeReviverMap;
    }
    return deepCopyByJSON;
}
const defaultPresetTypeReviverMap$1 = new Map(typeReviverArray);
/**
 * 通过 自定义 JSON 序列人的方式 对 value 进行 深拷贝
 */
const deepCopyByJSON = createDeepCopyByJSONWith(new Map(typeReviverArray));

exports.correctParse = correctParse;
exports.createCustomJSON = createCustomJSON;
exports.createCustomJSONParse = createCustomJSONParse;
exports.createCustomJSONStringify = createCustomJSONStringify;
exports.createDeepCopyByJSONWith = createDeepCopyByJSONWith;
exports.customJSONParse = customJSONParse;
exports.customJSONStringify = customJSONStringify;
exports.deepCopyByJSON = deepCopyByJSON;
exports.deepParse = deepParse;
exports.deepStringify = deepStringify;
exports.defaultPresetTypeReviverMap = defaultPresetTypeReviverMap$1;
exports.isJSONString = isJSONString;
exports.isReviverPair = isReviverPair;
exports.isTypeReviversPair = isTypeReviversPair;
exports.safelyParse = safelyParse;
exports.typeNameOfDefaultRevier = typeNameOfDefaultRevier;
